@@ -37,6 +37,7 @@ from yum.misc import checksum, _available_checksums, AutoFileChecksums
 from yum.repoMDObject import RepoMD, RepoMDError, RepoData
 from xml.dom import minidom
 from optparse import OptionParser
+from cStringIO import StringIO
 
 
 class RepoMetadata:
@@ -46,6 +47,9 @@ class RepoMetadata:
         self.repodir = os.path.abspath(repo)
         self.repomdxml = os.path.join(self.repodir, 'repomd.xml')
         self.compress_type = _available_compression[-1] # best available
+        self.compress = True
+        self.checksum_type = 'sha256'
+        self.unique_md_filenames = True
 
         if not os.path.exists(self.repomdxml):
             raise MDError, '%s not found' % self.repomdxml
@@ -103,6 +107,8 @@ class RepoMetadata:
         if isinstance(metadata, minidom.Document):
             md = metadata.toxml()
             mdname = 'updateinfo.xml'
+            oldmd = AutoFileChecksums(StringIO(md), [self.checksum_type])
+            oldmd.read()
         elif isinstance(metadata, str):
             if os.path.exists(metadata):
                 mdname = os.path.basename(metadata)
@@ -147,7 +153,7 @@ class RepoMetadata:
         new_rd.checksum = (self.checksum_type, csum)
         new_rd.size = str(os.stat(destmd).st_size)
         if self.compress:
-            new_rd.openchecksum = oldmd.checksums.hexdigests().popitem()
+            new_rd.openchecksum = (self.checksum_type, oldmd.checksums.hexdigests().popitem()[1])
             new_rd.opensize = str(oldmd.checksums.length)
         new_rd.timestamp = str(int(os.stat(destmd).st_mtime))
         self.repoobj.repoData[new_rd.type] = new_rd
@@ -236,7 +242,7 @@ def main(args):
     if opts.compress_type not in _available_compression:
         print "Compression %s not available: Please choose from: %s" % (opts.compress_type, ', '.join(_available_compression))
         return 1
-    if opts.sumtype not in _available_checksums:
+    if opts.sumtype != 'sha' and opts.sumtype not in _available_checksums:
         print "Checksum %s not available: Please choose from: %s" % (opts.sumtype, ', '.join(_available_checksums))
         return 1
     repomd.compress_type = opts.compress_type
